# ### sci-pkg ###
# store: functions that handle storage (object/cloud storage, posix file systems)
# definitions: 
#   Object: a chunk of information stored in an Object Store
#   File a chunk of information that touches file system
#   file_upload turns file into an object and file_download goes vice versa

import os

"""
Simplified classes for accessing object storage systems.
Currently only implemented for Swift Storage.

Current limitations: 
 - swift: object_put function limited to 5GB per object
"""
class swift:
    """
    Example:
        mystor = sci.store.swift('the-bucket', 'virtual/sub/dir') 
        my_objects = mystor.bucket_list()
    -------------------------
    Initialize an Openstack Swift container/bucket. 
    Data will be written to the root of the bucket unless the virtual dir
    prefix is set.The prefix can be a single virtual folder such as 
    'subfolder' or a virtual folder such as 'folder/subfolder/subfolder'.     
    For authentication you need to have set some environment 
    variables. This can either be OS_AUTH_URL, OS_TENANT_NAME, 
    OS_USERNAME, OS_PASSWORD or OS_STORAGE_URL and OS_AUTH_TOKEN. 
    If you have Swift Commander installed, a new auth token and
    storage URL will be automatically stored in the ~/.swift
    folder after you use the 'swc' command. 
    """

    def __init__(self, bucket, prefix=None):
        """ 
        Examples:
        mystor = swift('the-bucket', 'virtual/sub/directory')
        my_objects = mystor.bucket_list()
        my_filtered_objects = mystor.bucket_list({'proj': 'ABC', 'tag': 'new'})
        dict = mystor.object_put('the-object', "THE CONTENT", {tag': 'new'})
        obj = mystor.object_get('the-object')
        j = mystor.object_get_json('the-object.json')
        ret = mystor.object_meta_set('the-object', {'proj': 'XYZ'):
        dict = mystor.object_meta_get('the-object')
        """

        import swiftclient, json

        sw_auth_version = 2
        sw_authurl =  os.getenv("OS_AUTH_URL","")
        sw_user = os.getenv("OS_USERNAME", "")
        sw_key = os.getenv("OS_PASSWORD", "")
        self.tenant = os.getenv("OS_TENANT_NAME", "")
        self.storageurl = os.getenv("OS_STORAGE_URL","")
        self.authtoken = os.getenv("OS_AUTH_TOKEN","")
        if not sw_key and not self.authtoken:
            #reading authtoken from file cache in ~/.swift folder
            self.authtoken = self._get_set_token_file(sw_authurl)
        self.optsauth = self._get_swift_options(sw_authurl,sw_user,sw_key)

        self.swiftconn = swiftclient.client.Connection(authurl=sw_authurl, user=sw_user, key=sw_key,  
                auth_version=sw_auth_version, os_options=self.optsauth)
        try:
            self.storageurl, newauthtoken = self.swiftconn.get_auth()
        except swiftclient.exceptions.ClientException as e:
            print("Connection Error: %s" % e.msg)
            self._exiterr()

        self.bucket = bucket
        self.prefix = prefix

        if self.authtoken != newauthtoken:
            self.authtoken == newauthtoken
            # writing new authtoken to file cache in ~/.swift folder
            self._get_set_token_file(sw_authurl, newauthtoken)
            
        os.environ["OS_AUTH_TOKEN"] = self.authtoken
        self.optsauth['os_auth_token'] = self.authtoken
        self.optsauth['auth_token'] = self.authtoken
        os.environ["OS_STORAGE_URL"] = self.authtoken
        self.optsauth['os_storage_url'] = self.storageurl
        self.optsauth['object_storage_url'] = self.storageurl


    def bucket_list(self, filter=None):
        """
        return a list of objects from the previously initialized bucket. 
        Example:
        my_filtered_objects = mystor.bucket_list({'proj': 'ABC', 'tag': 'new'})
        -------------------------
        This list can be a filtered list by passing in a dictionary of
        key / value pairs as filter. 
        """

        if self.bucket == None:
            return []

        try:
            listing = self.swiftconn.get_container(self.bucket,
                                           prefix=self.prefix,
                                           full_listing=True)
        except swiftclient.exceptions.ClientException as e:
            print("Bucket List Error: %s (%s, %s)" % (e.msg, e.http_status,  e.http_reason))
            return []

        newlist=[]
        for item in listing[1]:
            if item['content_type'] != 'application/directory' and \
                not item['name'].startswith('.') and not '/.' in item['name']:
                    if not filter:
                        newlist.append(item["name"])
                    else:
                        meta = self.object_meta_get(item['name'])
                        # is filter a full subset of meta
                        if filter.items() <= meta.items():
                            newlist.append(item["name"])

        #return [item["name"] for item in listing[1]]
        return newlist

    def file_download(self, objnames, folder=None):
        """
        Download a list of objects to files. If the list has a single object
        the target "filepath" can be a file, otherwise the target will be a 
        local folder, if "filepath" is omitted we copy to the current folder.
        Example:
        c = mystor.file_download(['prefix/f.txt'], 'fld/fld2') or
        c = mystor.file_download(['f.txt','g.txt'], 'fld') or
        c = mystor.file_download(['f.txt','g.txt']) or
        -------------------------
        returns a list of downloaded files or an empty list in a case of error
        """
        # see https://docs.openstack.org/python-swiftclient/latest/service-api.html

        import swiftclient.service 

        if self.bucket == None:
            return []

        objnames = self._fix_object_path(objnames)

        startdir= os.getcwd()
        if folder:
            folder = os.path.expanduser(folder)
            if not os.path.exists(folder):
                print('folder does not exist: %s' % folder)
                return False
            else:
                os.chdir(folder)

        currdir=os.getcwd()
        downloaded_files = []
        with swiftclient.service.SwiftService(options=self.optsauth) as sw:
            try:
                print("Downloading to '%s' ..." % currdir)
                for r in sw.download(
                        container=self.bucket,
                        objects=objnames):
                    if r['success']:
                        if 'object' in r and 'path' in r:
                            print("'%s' downloaded to '%s'" % (r['object'],r['path']))
                            downloaded_files.append(os.path.join(currdir,r['path']))
                    else:
                        if 'object' in r and 'path' in r:
                            print("'%s' download to '%s' failed" % (r['object'],r['path']))
                        if 'error' in r and 'attempts' in r:
                            print("  Error: '%s', attempts: %s" % (r['error'],r['attempts']))

            except swiftclient.service.SwiftError as e:
                print("Error:", e.value)

        os.chdir(startdir)
        return downloaded_files

    def file_upload(self, filepaths, objname=None, metadict=None):
        """
        Upload a list of files. If the list has a single entry it will be 
        copied to the target "prefix/objname", otherwise objname is ignored.
        if "objname" is omitted, the object is called prefix/filename. 
        Example:
        c = mystor.file_upload(['fld/f.txt'], 'x.txt') or
        c = mystor.file_upload(['f.txt','g.txt'])
        ------------------------- 
        returns a list of uploaded objects or an empty list in a case of error
        """
        # see https://docs.openstack.org/python-swiftclient/latest/service-api.html
        import swiftclient.service, getpass

        if self.bucket == None:
            return []

        uploaded_objects = []
        with swiftclient.service.SwiftService(options=self.optsauth) as sw:

            objs = self._fix_file_paths(filepaths)
            objs = [
                swiftclient.service.SwiftUploadObject(
                    #o, object_name=o.replace(
                    #dir, 'my-%s-objects' % dir, 1)
                    o, object_name=self._upload_object_name(o)
                ) for o in objs
            ]
            # if there is only 1 file and target objname is set
            if objname and len(objs) == 1:
                objs[0].object_name=self._fix_object_path([objname])[0]

            #for o in objs:
            #    print("***",o.object_name,o.source)

            for r in sw.upload(container=self.bucket, 
                objects=objs, 
                options={'segment_size':104857600,
                        'use_slo':True,
                        'meta': {'uploaded-by': getpass.getuser()},
                        'segment_container':'.segments_'+self.bucket,
                        'shuffle': True}):
                if r['success']:
                    if 'object' in r and 'status' in r:
                        print("object '%s' %s." % (r['object'],r['status']))
                        uploaded_objects.append(r['object'])
                else:
                    if 'object' in r:
                        print("object '%s' upload failed" % r['object'])
                    if 'error' in r:
                        print("  Error: '%s'" % r['error'])
        return uploaded_objects


    def object_get(self, objname):
        """
        Load the object into memory
        Example:
        content = mystor.object_get('prefix/myobj.json') or
        content = mystor.object_get('myobj.json') or
        -------------------------
        using a prefix path is optional and if no prefix 
        is added we prepend the default prefix 
        for tools using file handles (such as pandas) use 
        io.StringIO or io.BytesIO :
        handle = io.StringIO(content.decode('utf-8')) 
           # or handle = io.BytesIO(content)
        dataframe = pd.read_csv(handle)
        """
        if self.bucket == None:
            return None

        objname = self._fix_object_path([objname])[0]
        
        content = self.swiftconn.get_object(self.bucket, objname)[1]
        return content 

    def object_get_json(self, objname):
        """ 
        load object into memory and de-serialize json 
        Example:
        j = mystor.object_get_json('prefix/myobj.json')
        print(json.dumps(j, indent=2))
        """
        import json
        content = self.object_get(objname)
        return json.loads(content.decode('utf-8'))

    def object_get_csv(self, objname, dictreader=True, dialect=None):
        """ 
        load swift object into memory and de-serialize csv 
        Examples:
        -----------------------------------------------------------------
        table = mystor.object_get_csv('prefix/myobj.csv')
        for row in table:
            for field in table.fieldnames:
                print(field,':',row[field])
        -------------------------
        table = mystor.object_get_csv('toolbox/pi_all.csv',False,'excel')
        headers = next(table, None)
        print(headers)
        for row in table:
            print (row)
        -------------------------
        By default the function returns csv.DictReader object.
        The DictReader object has the attribute 'fieldnames' that returns
        the csv header as a list. Optionally you can use csv.reader 
        instead of csv.DictReader and a different dialect, such as 'excel'.
        If you do not want to use the internal python csv package you can
        use the io.StringIO function to create an io handle in memory that
        can be used instead of a file handle, e.g. for pandas:
        -------------------------
        content= self.object_get(objname)
        handle = io.StringIO(content.decode('utf-8')) 
           # or handle = io.BytesIO(content)
        dataframe = pd.read_csv(handle)
        """
        import io, csv
        content= self.object_get(objname)
        if not content:
            return None
        handle = io.StringIO(content.decode('utf-8'))
        # sniffer does not seem to work here
        #dialect = csv.Sniffer().sniff(handle.read(4096))
        if dictreader:
            if dialect:
                table = csv.DictReader(handle, dialect)
            else:
                table = csv.DictReader(handle)
        else:
            if dialect:
                table = csv.reader(handle, dialect)
            else:
                table = csv.reader(handle)
        return table

    def object_meta_get(self, objname):
        """ 
        retrieve custom metadata from object as a dictionary
        Example: 
        dict = mystor.object_meta_get('myobj.json')
        -------------------------
        The function will return a dictionary of metadata 
        """
        if self.bucket == None:
            return None

        objname = self._fix_object_path([objname])[0]
        head = self.swiftconn.head_object(self.bucket,objname)
        newhead = {}
        for k,v in head.items():
            if k.startswith('x-object-meta-'):
                newhead[k.replace('x-object-meta-','')] = v
        return newhead

    def object_meta_set(self, objname, metadict):
        """ 
        set custom metadata on existing object using a dictionary 
        Example:
        dict = mystor.object_meta_set('myobj.json', {'key': 'val', 'x': 'y'})
        -------------------------
        The function will return ????
        """
        if self.bucket == None:
            return None
        objname = self._fix_object_path([objname])[0]
        metadict = self._fix_metadict(metadict)

        resp = dict()
        #self.swiftconn.post_object(self.storageurl, token=self.authtoken, \
        #    bucket=self.bucket, name="%s/%s" % (self.prefix,objname), \
        #    headers=metadict, http_conn=None, response_dict=resp, service_token=None)
        ret = self.swiftconn.post_object(self.bucket, objname, \
            headers=metadict, response_dict=resp)
        return resp

    def object_put(self, objname, content, metadict=None):
        """ 
        save object to bucket under prefix and optionally set metadata dict.
        Example:
        mystor.object_put('myobj.dat', x, {'key': 'val', 'a': 'b'})

        """

        if self.bucket == None:
            return None
        metadict = self._fix_metadict(metadict)

        #self.swiftconn.put_object(self.storageurl, \
        #    bucket=self.bucket, name="%s/%s" % (self.prefix,objname), \
        #    contents=content, content_length=None, etag=None, chunk_size=None, \
        #    content_type=None, headers=metadict, http_conn=None, proxy=None, \
        #    query_string=None, response_dict=None, service_token=None)

        objname = self._fix_object_path([objname])[0]

        resp = dict()
        ret = self.swiftconn.put_object(self.bucket, objname, content, \
            response_dict=resp, headers=metadict)
        return resp

    def object_put_json(self, objname, content, metadict=None):
        """ 
        save json object to bucket and optionally set metadata using a dict
        Example:
        mystor.object_put('myobj.json', x, {'key': 'val', 'a': 'b'})
        """

        import json

        if self.bucket == None:
            return None        
        j = json.dumps(content, indent=4)
        return self.object_put(objname,j,metadict)

    def object_put_pickle(self, objname, content, metadict=None):
        """ 
        save pickle object to bucket and optionally set metadata using a dict. 
        Example:
        mystor.object_put_pickle('myobj.dat', x, {'key': 'val', 'a': 'b'})
        Large Pickle objects can be faster than large json objects, however
        they are proprietary Python data objects and cannot be used by R, etc.  
        """
        import pickle

        if self.bucket == None:
            return None        
        p = pickle.dumps(content)
        return self.object_put(objname,p,metadict)

    def object_put_csv(self, objname, content, metadict=None, dictwriter=False, dialect=None):
        """ 
        save csv object to bucket and optionally set metadata using a dict
        Example:
        mystor.object_put_csv('myobj.csv', x, {'key': 'val', 'a': 'b'})
        """
        import io, csv

        if self.bucket == None:
            return None

        handle = io.StringIO()
        # sniffer does not seem to work here
        #dialect = csv.Sniffer().sniff(handle.read(4096))
        if dictwriter:
            if dialect:
                wr = csv.DictWriter(handle, dialect)
                wr.writerows(content)
            else:
                wr = csv.DictWriter(handle)
                wr.writerows(content)
        else:
            if dialect:
                wr = csv.writer(handle, dialect)
                wr.writerows(content)
            else:
                wr = csv.writer(handle)
                wr.writerows(content)

        buffer = handle.getvalue()

        return self.object_put(objname,buffer,metadict)

    def _object_ext(self, name):
        """ get the object extention (like content type) """
        ext = name.split(".")[-1]
        if ext == name:
            return ""
        return ext

    def _exiterr(self):
        import sys
        sys.exit(1)
    
    def _fix_object_path(self, objnames):
        """ 
        add object prefix to list of objects if needed 
        """        
        newobjs=[]
        for o in objnames:
            #print('######## o type', type(o), 0)
            if o.find('/') > -1:
                obj = o
            else:
                if not self.prefix:
                    obj = o
                else:
                    obj = "%s/%s" % (self.prefix,o)
            newobjs.append(obj)                
        return newobjs

    def _upload_object_name(self, objname):
        """ 
        add object prefix to single upload object
        """        
        if objname.startswith(self.prefix+'/'):
            o = objname
        elif os.path.isabs(objname):
            o=self.prefix+'/'+os.path.basename(objname)
        else:
            o=self.prefix+'/'+objname
        return o.replace('//','/')

    def _fix_file_paths(self, files):
        """ 
        changing back to forward slashes (for windows)
        """
        newfiles=[]
        for f in files:
            newfiles.append(f.replace('\\','/'))
        return newfiles

    def _fix_metadict(self, metadict):
        """ ensure that x-object-meta prefix is added to all metadata keys """
        if not metadict:
            return metadict
        metadict2 = {}
        for k,v in metadict.items():
            if k.startswith('x-object-meta-'):
                j = k
            else:
                j = "x-object-meta-%s" % k
            metadict2[j] = v
        return metadict2

    def _get_swift_options(self, sw_authurl, sw_user, sw_key):

        if sw_key:
            if not sw_authurl:
                print ("Please set env var OS_AUTH_URL, e.g to https://host.domain.org/auth/v2.0")
                return {}
            if not sw_user:
                print ("Please set environment variable OS_USERNAME, e.g. your user name")
                return {}
            if not self.tenant:
                print ("Please set environment variable OS_TENANT_NAME, e.g. AUTH_tenantname")
                return {}
            print('  using password to authenticate ...')
            options = {
                "auth_version": "2.0",
                "tenant_name": self.tenant,
                "os_tenant_name": self.tenant,

                # or service_type, endpoint_type, tenant_name, object_storage_url, 
                # region_name, service_username, service_project_name, service_key
            }
        elif self.authtoken:
            if not self.tenant:
                print ("Please set environment variable OS_TENANT_NAME")
                return {}
            if not self.storageurl:
                print ("Please set environment variable OS_STORAGE_URL")
                return {}
            options = {
                "auth_version": "2.0",
                "tenant_name" : self.tenant,
                "os_tenant_name": self.tenant,
                "object_storage_url" : self.storageurl,
                "auth_token" : self.authtoken,
            }
        else:
            options = {}
            print('Please set either OS_AUTH_TOKEN or OS_PASSWORD environment ' \
                  'variables. (e.g. execute "read -s OS_PASSWORD")')
        return options

    def _get_set_token_file(self, authurl, newauthtoken=None):
        import urllib.parse
        homedir = os.path.expanduser("~")        
        host = urllib.parse.urlparse(authurl).netloc
        authtokenfile = os.path.join(homedir,'.swift','auth_token_%s_v2_%s' % (host,self.tenant))
        storageurlfile = os.path.join(homedir,'.swift','storageurl_%s_v2_%s' % (host,self.tenant))

        if newauthtoken:
            if not os.path.exists(os.path.join(homedir,'.swift')):
                os.mkdir(os.path.join(homedir,'.swift'))
            if os.path.exists(authtokenfile):
                os.remove(authtokenfile)
            with open(authtokenfile, 'w') as f:
                f.write(newauthtoken)
            with open(storageurlfile, 'w') as f:
                f.write(self.storageurl)
            return ""
        else:
            if os.path.exists(authtokenfile):
                with open(authtokenfile, 'r') as f:
                    self.authtoken = f.readline().strip()
                with open(storageurlfile, 'r') as f:
                    self.storageurl = f.readline().strip()
                return self.authtoken
            else:
                return ""



class s3:
    """
    Initialize an s3 bucket using a certain profile
    Data will be written to the root of the bucket unless the virtual dir
    prefix is set.The prefix can be a single virtual folder such as 
    'subfolder' or a virtual folder such as 'folder/subfolder/subfolder'.     
    Example:
        mystor = sci.store.s3('the-bucket', 'virtual/dir', 'profile') 
        my_objects = mystor.bucket_list()
    """

    def __init__(self, bucket, prefix=None, profile='default'):
        """ 
        Examples:
        mystor = s3('the-bucket', 'virtual/sub/directory')
        """
        self.s3conn = "s3"
        self.bucket = bucket
        self.prefix = prefix

    def bucket_list(self, filter=None):
        """
        return a list of objects from the previously initialized bucket. 
        This list can be a filtered list by passing in a dictionary of
        key / value pairs as filter. 

        Example:
        my_filtered_objects = mystor.bucket_list({'proj': 'ABC', 'tag': 'new'})
        """

        if self.bucket == None:
            return []

        try:
            listing = self.s3conn.get_container(self.bucket,
                                           prefix=self.prefix,
                                           full_listing=True)
        except:
            return []

        newlist=[]
        for item in listing[1]:
            if item['content_type'] != 'application/directory' and \
                not item['name'].startswith('.') and not '/.' in item['name']:
                    if not filter:
                        newlist.append(item["name"])
                    else:
                        meta = self.object_meta_get(item['name'])
                        # is filter a full subset of meta
                        if filter.items() <= meta.items():
                            newlist.append(item["name"])

        #return [item["name"] for item in listing[1]]
        return newlist

    def object_get(self, objname):
        """ load the object into memory """
        if self.bucket == None:
            return None
        raw = self.swiftconn.get_object(self.bucket, objname)[1]
        return raw 
        #raw = lzma.decompress(compressed)
        #data = json.loads(raw.decode("utf-8"))
        # do something with the data ...

    def object_get_json(self, objname):
        """ load object into memory and de-serialize json """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_get_csv(self, objname):
        """ load object into memory and convert to csv """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_meta_get(self, objname):
        """ retrieve custom metadata from object as a dictionary """
        if self.bucket == None:
            return None
        head = self.s3conn.head_object(self.bucket,objname)
 
    def object_meta_set(self, objname, metadict):
        """ set custom metadata on existing object using a dictionary """

        if self.bucket == None:
            return None

        ret = self.s3conn.post_object(self.bucket, "%s/%s" % (self.prefix,objname))        

    def object_put(self, objname, content, metadict=None):
        """ save object to bucket and optionally set metadata using a dict """

        if self.bucket == None:
            return None

        ret = self.s3conn.put_object(self.bucket, "%s/%s" % (self.prefix,objname), content, \
            response_dict=resp, headers=metadict)

    def _object_ext(self, name):
        """ get the object extention (like content type) """
        ext = name.split(".")[-1]
        if ext == name:
            return ""
        return ext


class google:
    """
    Initialize a Google cloud bucket using a certain profile
    Data will be written to the root of the bucket unless the virtual dir
    prefix is set.The prefix can be a single virtual folder such as 
    'subfolder' or a virtual folder such as 'folder/subfolder/subfolder'.     
    Example:
        mystor = sci.store.google('the-bucket', 'virtual/dir', 'profile') 
        my_objects = mystor.bucket_list()
    """

    def __init__(self, bucket, prefix=None, profile='default'):
        """ 
        Examples:
        mystor = s3('the-bucket', 'virtual/sub/directory')
        """
        self.s3conn = "s3"
        self.bucket = bucket
        self.prefix = prefix

    def bucket_list(self, filter=None):
        """
        return a list of objects from the previously initialized bucket. 
        This list can be a filtered list by passing in a dictionary of
        key / value pairs as filter. 

        Example:
        my_filtered_objects = mystor.bucket_list({'proj': 'ABC', 'tag': 'new'})
        """

        if self.bucket == None:
            return []

        try:
            listing = self.s3conn.get_container(self.bucket,
                                           prefix=self.prefix,
                                           full_listing=True)
        except:
            return []

        newlist=[]
        for item in listing[1]:
            if item['content_type'] != 'application/directory' and \
                not item['name'].startswith('.') and not '/.' in item['name']:
                    if not filter:
                        newlist.append(item["name"])
                    else:
                        meta = self.object_meta_get(item['name'])
                        # is filter a full subset of meta
                        if filter.items() <= meta.items():
                            newlist.append(item["name"])

        #return [item["name"] for item in listing[1]]
        return newlist

    def object_get(self, objname):
        """ load the object into memory """
        if self.bucket == None:
            return None
        raw = self.swiftconn.get_object(self.bucket, objname)[1]
        return raw 
        #raw = lzma.decompress(compressed)
        #data = json.loads(raw.decode("utf-8"))
        # do something with the data ...

    def object_get_json(self, objname):
        """ load object into memory and de-serialize json """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_get_csv(self, objname):
        """ load object into memory and convert to csv """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_meta_get(self, objname):
        """ retrieve custom metadata from object as a dictionary """
        if self.bucket == None:
            return None
        head = self.s3conn.head_object(self.bucket,objname)
 
    def object_meta_set(self, objname, metadict):
        """ set custom metadata on existing object using a dictionary """

        if self.bucket == None:
            return None

        ret = self.s3conn.post_object(self.bucket, "%s/%s" % (self.prefix,objname))        

    def object_put(self, objname, content, metadict=None):
        """ save object to bucket and optionally set metadata using a dict """

        if self.bucket == None:
            return None

        ret = self.s3conn.put_object(self.bucket, "%s/%s" % (self.prefix,objname), content, \
            response_dict=resp, headers=metadict)

    def _object_ext(self, name):
        """ get the object extention (like content type) """
        ext = name.split(".")[-1]
        if ext == name:
            return ""
        return ext


 
class azure:
    """
    Initialize an s3 bucket using a certain profile
    Data will be written to the root of the bucket unless the virtual dir
    prefix is set.The prefix can be a single virtual folder such as 
    'subfolder' or a virtual folder such as 'folder/subfolder/subfolder'.     
    Example:
        mystor = sci.store.azure('the-bucket', 'virtual/dir', 'profile') 
        my_objects = mystor.bucket_list()
    """

    def __init__(self, bucket, prefix=None, profile='default'):
        """ 
        Examples:
        mystor = s3('the-bucket', 'virtual/sub/directory')
        """
        self.s3conn = "s3"
        self.bucket = bucket
        self.prefix = prefix

    def bucket_list(self, filter=None):
        """
        return a list of objects from the previously initialized bucket. 
        This list can be a filtered list by passing in a dictionary of
        key / value pairs as filter. 

        Example:
        my_filtered_objects = mystor.bucket_list({'proj': 'ABC', 'tag': 'new'})
        """

        if self.bucket == None:
            return []

        try:
            listing = self.s3conn.get_container(self.bucket,
                                           prefix=self.prefix,
                                           full_listing=True)
        except:
            return []

        newlist=[]
        for item in listing[1]:
            if item['content_type'] != 'application/directory' and \
                not item['name'].startswith('.') and not '/.' in item['name']:
                    if not filter:
                        newlist.append(item["name"])
                    else:
                        meta = self.object_meta_get(item['name'])
                        # is filter a full subset of meta
                        if filter.items() <= meta.items():
                            newlist.append(item["name"])

        #return [item["name"] for item in listing[1]]
        return newlist

    def object_get(self, objname):
        """ load the object into memory """
        if self.bucket == None:
            return None
        raw = self.swiftconn.get_object(self.bucket, objname)[1]
        return raw 
        #raw = lzma.decompress(compressed)
        #data = json.loads(raw.decode("utf-8"))
        # do something with the data ...

    def object_get_json(self, objname):
        """ load object into memory and de-serialize json """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_get_csv(self, objname):
        """ load object into memory and convert to csv """
        raw = self.object_get(objname)
        return json.loads(raw.decode()) #json.loads(raw.decode("utf-8"))

    def object_meta_get(self, objname):
        """ retrieve custom metadata from object as a dictionary """
        if self.bucket == None:
            return None
        head = self.s3conn.head_object(self.bucket,objname)
 
    def object_meta_set(self, objname, metadict):
        """ set custom metadata on existing object using a dictionary """

        if self.bucket == None:
            return None

        ret = self.s3conn.post_object(self.bucket, "%s/%s" % (self.prefix,objname))        

    def object_put(self, objname, content, metadict=None):
        """ save object to bucket and optionally set metadata using a dict """

        if self.bucket == None:
            return None

        ret = self.s3conn.put_object(self.bucket, "%s/%s" % (self.prefix,objname), content, \
            response_dict=resp, headers=metadict)

    def _object_ext(self, name):
        """ get the object extention (like content type) """
        ext = name.split(".")[-1]
        if ext == name:
            return ""
        return ext



# ### sci-pkg ###
# db: functions for DBMS (postgres, mysql, sqlite, mssql etc)
#

import os

class postgres:
    """
    Examples:
    mydb = mssql('the-bucket', 'virtual/sub/directory')
    
    the password should be stored in ~/.pgpass
    """
    def __init__(self, host, dbname, user, password, port=5432):
        """ 
        just pass in a bucket and an optional prefix. The prefix can be a 
        single virtual folder such as 'subfolder' or a virtual folder 
        such as 'folder/subfolder/subfolder'. Example:
        mystor = SwiftStorage('the-bucket', 'virtual/sub/directory') 
        """
        import psycopg2
    def execute(self, sqlstr):
        """
        execute a sql command on the database
        """
    def fetch(self, sqlstr):
        """
        return the result of a SQL query as a dictionary
        """

class mysql:
    """
    Examples:
    mystor = mssql('the-bucket', 'virtual/sub/directory')
    """
    def __init__(self, host, dbname, user, password, port=3306):
        """ 
        just pass in a bucket and an optional prefix. The prefix can be a 
        single virtual folder such as 'subfolder' or a virtual folder 
        such as 'folder/subfolder/subfolder'. Example:
        mystor = SwiftStorage('the-bucket', 'virtual/sub/directory') 
        """
        import PyMySQL
    def execute(self, sqlstr):
        """
        execute a sql command on the database
        """
    def fetch(self, sqlstr):
        """
        return the result of a SQL query as a dictionary
        """

class mssql:
    """
    Examples:
    mystor = mssql('the-bucket', 'virtual/sub/directory')
    """
    def __init__(self, host, dbname, user, password, port=1433):
        """ 
        just pass in a bucket and an optional prefix. The prefix can be a 
        single virtual folder such as 'subfolder' or a virtual folder 
        such as 'folder/subfolder/subfolder'. Example:
        mystor = SwiftStorage('the-bucket', 'virtual/sub/directory') 
        """
        import pyodbc
    def execute(self, sqlstr):
        """
        execute a sql command on the database
        """
    def fetch(self, sqlstr):
        """
        return the result of a SQL query as a dictionary
        """

class sqlite:
    """
    Examples:
    mystor = mssql('the-bucket', 'virtual/sub/directory')
    """
    def __init__(self, dbname):
        """ 
        dbname is a file name of a local database. 
        use :memory: as filename to create the database in RAM 
        """
        import sqlite3
    def execute(self, sqlstr):
        """
        execute a sql command on the database
        """
    def fetch(self, sqlstr):
        """
        return the result of a SQL query as a dictionary
        """

def _getdbpasswd(host, port, dbname, user):
    passwd = os.path.join(os.path.expanduser("~"),'.sci','.dbpasswd')
    if not os.path.exists(passwd):
        return None 
    with open(passwd, 'r') as f:
        host1,port1,dbname1,user1,password1 = f.readline().strip().split(':')




# ### sci-pkg ###
# tools: commonly used helper functions
#

import os, socket

def logger(name=None, stderr=False):
    """ eg: loghandle = sci.tools.logger('loggername')
    """
    import logging, logging.handlers
    # levels: CRITICAL:50,ERROR:40,WARNING:30,INFO:20,DEBUG:10,NOTSET:0
    if not name:
        name=__file__.split('/')[-1:][0]
    l=logging.getLogger(name)
    l.setLevel(logging.INFO)
    f=logging.Formatter('%(name)s: %(levelname)s:%(module)s.%(lineno)d: %(message)s')
    # logging to syslog
    s=logging.handlers.SysLogHandler('/dev/log')
    s.formatter = f
    l.addHandler(s)
    if stderr:
        l.setLevel(logging.DEBUG)
        # logging to stderr        
        c=logging.StreamHandler()
        c.formatter = f
        l.addHandler(c)
    return l

def uniq(seq):
    """ remove duplicates from a list """ 
    # Not order preserving
    keys = {}
    for e in seq:
        keys[e] = 1
    return list(keys.keys())

def getcsv(url):
    """ read a large csv file from a web server"""
    import requests, contextlib, csv

    with contextlib.closing(requests.get(url, stream=True)) as r:
        f = (line.decode('utf-8') for line in r.iter_lines())
        reader = csv.reader(f, delimiter=',', quotechar='"')
        for row in reader:
            print(row)

def most_common(seq):
    """ eg: most = most_common(['A','B',B','C'])
        Which is the most commonly occuring element in a sequence (e.g. list)
        In this case: 'B'
    """
    import itertools
    # get an iterable of (item, iterable) pairs
    SL = sorted((x, i) for i, x in enumerate(seq))
    # print 'SL:', SL
    groups = itertools.groupby(SL, key=operator.itemgetter(0))
    # auxiliary function to get "quality" for an item
    def _auxfun(g):
        item, iterable = g
        count = 0
        min_index = len(seq)
        for _, where in iterable:
            count += 1
            min_index = min(min_index, where)
            # print 'item %r, count %r, minind %r' % (item, count, min_index)
        return count, -min_index
        # pick the highest-count/earliest item
    return max(groups, key=_auxfun)[0]


def uid2user(uid):
    """ eg: user = uid2user(54321)
        get username based on uidnumber, return uidNumber if fails
    """
    try:
        import pwd
        return pwd.getpwuid(uid).pw_name
    except:
        return str(uid)

def mail_status(to, subject, text, attachments=[], cc=[], bcc=[], smtphost="", fromaddr=""):
    """ example: ret = mail_status(['john@doe.org'], "this subject", "that body")

        every email body starts with this content. $notify_text is substituted with "text":
            'This is a notification message from $application, running on \n' + \
            'host $host. Please review the following message:\n\n' + \
            '$notify_text\n\nIf output is being captured, you may find additional\n' + \
            'information in your logs.\n'
    """
    from string import Template

    body = Template('This is a notification message from $application, running on \n' + \
            'host $host. Please review the following message:\n\n' + \
            '$notify_text\n\nIf output is being captured, you may find additional\n' + \
            'information in your logs.\n'
            )
    host_name = socket.gethostname()
    full_body = body.substitute(host=host_name.upper(), notify_text=text, application=os.path.basename(__file__))

    return mail_send(to,subject,full_body,attachments,cc,bcc,smtphost,fromaddr) 

def mail_send(to, subject, body, attachments=(), cc=(), bcc=(), smtphost="", fromaddr=""):
    """ example: ret = mail_send(['john@doe.org'], 'this subject', 'that body',['file1','file2'])"""

    import smtplib, socket
    from email.mime.multipart import MIMEMultipart
    from email.mime.base import MIMEBase
    from email.mime.text import MIMEText
    from email.utils import COMMASPACE, formatdate
    from email import encoders as Encoders

    if not isinstance(to,list):
        print("the 'to' parameter needs to be a list")
        return False    
    if len(to)==0:
        print("no 'to' email addresses")
        return False

    myhost=socket.getfqdn()

    if smtphost == '':
        smtphost = _get_mx_from_email_or_fqdn(myhost)
    if not smtphost:
        sys.stderr.write('could not determine smtp mail host !\n')

    if fromaddr == '':
        fromaddr = os.path.basename(__file__) + '-no-reply@' + \
           '.'.join(myhost.split(".")[-2:]) #extract domain from host
    tc=0
    for t in to:
        if '@' not in t:
            # if no email domain given use domain from local host
            to[tc]=t + '@' + '.'.join(myhost.split(".")[-2:])
        tc+=1

    message = MIMEMultipart()
    message['From'] = fromaddr
    message['To'] = COMMASPACE.join(to)
    message['Date'] = formatdate(localtime=True)
    message['Subject'] = subject
    message['Cc'] = COMMASPACE.join(cc)
    message['Bcc'] = COMMASPACE.join(bcc)

    if '<html>' or '<body>' in body.lower():
        message.attach(MIMEText(body, 'html'))
    else:
        message.attach(MIMEText(body, 'plain'))

    for f in attachments:
        part = MIMEBase('application', 'octet-stream')
        part.set_payload(open(f, 'rb').read())
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        message.attach(part)

    addresses = []
    for x in to:
        addresses.append(x)
    for x in cc:
        addresses.append(x)
    for x in bcc:
        addresses.append(x)

    try:
        smtp = smtplib.SMTP(smtphost)
        smtp.sendmail(fromaddr, addresses, message.as_string())
        smtp.close()
        return True
    except smtplib.SMTPException as err:
        sys.stderr.write(str(err))
        sys.stderr.write('\n')
        return False
    
    
def _get_mx_from_email_or_fqdn(addr):
    """retrieve the first mail exchanger dns name from an email address."""
    import re
    # Match the mail exchanger line in nslookup output.
    MX = re.compile(r'^.*\s+mail exchanger = (?P<priority>\d+) (?P<host>\S+)\s*$')
    # Find mail exchanger of this email address or the current host
    if '@' in addr:
        domain = addr.rsplit('@', 2)[1]
    else:
        domain = '.'.join(addr.rsplit('.')[-2:])
    try:
        p = os.popen('/usr/bin/nslookup -q=mx %s' % domain, 'r')
    except:
        return ''
    mxes = list()
    for line in p:
        m = MX.match(line)
        if m is not None:
            mxes.append(m.group('host')[:-1])  #[:-1] just strips the ending dot
    if len(mxes) == 0:
        return ''
    else:
        return mxes[0]



import json, requests


# ### sci-pkg ###
# fh: Fred Hutch specific helper functions
#     (e.g. toolbox and pubmed stuff)


def getToolbox(jsonfile):
    """
    loads a jsonfile from Toolbox into a list of dictionaries 
    or a simple dictionary.
    """
    if not jsonfile:
        print("You need to pass JSON file name from toolbox")
        return ""
    try:
        r = requests.get("https://toolbox.fhcrc.org/json/%s" % jsonfile)
        r.raise_for_status()
        return r.json()
    except requests.exceptions.HTTPError as err:
        print("Error: {}".format(err))
        return ""


###### interaction with Pubmed ############################


def getPapers(j, papers, pi, lastname, forename, initial, sinceyear, prn=True):
    if not papers:
        return []
    k = 0
    rank = 0
    month = ""
    year = ""
    aid = ""
    ISSN = ""
    rows = []
    for i, paper in enumerate(papers["PubmedArticle"]):
        article = paper["MedlineCitation"]["Article"]
        journal = article["Journal"]
        authors = article["AuthorList"]
        aids = article["ELocationID"]
        journalinfo = paper["MedlineCitation"]["MedlineJournalInfo"]

        retlst = []
        if len(authors) > 0:
            rank, author = authorRank(authors, lastname, forename, initial)

        # if rank == 0 or (rank > 1 and rank < len(authors)-3):
        # only show pubs with first and last author
        #    continue
        # if len(aids)==0:
        #    continue
        if "Month" in journal["JournalIssue"]["PubDate"]:
            month = _month2num(journal["JournalIssue"]["PubDate"]["Month"])
            year = journal["JournalIssue"]["PubDate"]["Year"]

        if year == "" or year < sinceyear:
            continue

        k += 1
        if prn:
            print("%d) %s" % (k, article["ArticleTitle"]))

        if "ISSNLinking" in journalinfo:
            ISSN = journalinfo["ISSNLinking"]
        if len(aids) > 0:
            aid = aids[0]
            if prn:
                print("    ID: %s" % aids[0])

        first_or_last = 0
        if rank == 1:
            first_or_last = 1
        elif rank == len(authors):
            first_or_last = 1

        if prn:
            print("    Journal: %s, ISSN: %s" % (journal["Title"], ISSN))
            print("    Year: %s, Month: %s" % (year, month))
            print("    # of Authors: ", len(authors))
            print("    %s rank: %s" % (author, rank))
            print("    First or Last: %s" % first_or_last)

        peers = getPeers(j, authors, lastname, forename, initial)
        if peers:
            if prn:
                print("    PEERS ***:", ", ".join(peers))

        # if 'GrantList' in article:
        # ret = getGrants(article['GrantList'])

        # returns pi, author, aid, title, year, month, authorcount, authorpos, first_or_last, journal, issn
        first_or_last = 0
        if rank == 1:
            first_or_last = 1
        elif rank == len(authors):
            first_or_last = 1

        rows.append(
            [
                pi,
                author,
                aid,
                article["ArticleTitle"].encode("ascii", "ignore").decode(),
                year,
                month,
                len(authors),
                rank,
                first_or_last,
                len(peers),
                ";".join(peers),
                journal["Title"],
                ISSN,
            ]
        )

    return rows


def getGrants(grants):
    ### this is just a placeholder for now """
    for g in grants:
        if "GrantID" in g:
            print("   ", g["GrantID"], "(", g["Agency"], ")")


def authorRank(authors, lastname, forename, initials):
    """ get the position of a person in an author list, first do first name
        matching, then in a second loop try matching of initials.
    """
    i = 0
    retlist = []
    for a in authors:
        i += 1
        # if a['AffiliationInfo']:
        #    print ("Affiliation: ", a['AffiliationInfo'][0]['Affiliation'])
        if "LastName" in a and lastname:
            if "ForeName" in a:
                # print ("FORENAME", a['ForeName'], forename)
                if a["LastName"].lower() == lastname.lower() and (
                    forename.lower() in a["ForeName"].lower()
                    or a["ForeName"].lower() in forename.lower()
                ):
                    # retlist = [i, a['LastName'] + ' ' + a['ForeName'] + ' (' + a['Initials'] + ')']
                    retlist = [i, a["LastName"] + " " + a["ForeName"]]

    if len(retlist) > 0:
        return retlist
    i = 0
    for a in authors:
        i += 1
        if "LastName" in a:
            if "Initials" in a:
                # print("LAST", a['LastName'], lastname)
                if (
                    a["LastName"].lower() == lastname.lower()
                    and initials.lower() in a["Initials"].lower()
                ):
                    retlist = [i, a["LastName"] + " " + a["Initials"]]

    if len(retlist) > 0:
        return retlist
    else:
        return 0, ""


def getPeers(j, authors, lastname, forename, initial, institute="Fred Hutch"):
    """ get peer faculty from same institution, affiliation info in publication needs
        to include string institute, e.g. Fred Hutch'
    """
    import sci

    peers = []
    if not j:
        return peers
    HutchPIs = sci.tools.uniq(sci.json.jget(j, "displayName"))
    for a in authors:
        if not "LastName" in a:
            continue
        if not "ForeName" in a:
            continue
        if not "Initials" in a:
            continue
        if not "AffiliationInfo" in a:
            continue
        affinfo = a["AffiliationInfo"]
        aff = ""
        if len(affinfo) > 0:
            aff = affinfo[0]["Affiliation"]
        if not institute in aff:
            continue
        for hutchpi in HutchPIs:
            hlast, hfore, hinitial, hinitials = splitName(hutchpi)
            # skip current authorn
            if hlast.lower() == lastname.lower():
                # print ("***hlast, hfore, hinit:", hlast, hfore, hinit)
                # print ("***lastname, forename, initials:", lastname, forename, initials)
                if (
                    hfore.lower() in forename.lower()
                    or forename.lower() in hfore.lower()
                ):
                    continue
                if hinitial.lower() == initial.lower():
                    continue

            if a["LastName"].lower() == hlast.lower():
                if a["Initials"][0].lower() == hinitial.lower():
                    peers.append(hlast + " " + hfore)
    return peers


def splitName(fullname):
    fullname = fullname.replace(",", " ")
    fullname = fullname.replace(".", "")
    name = fullname.split()
    lastname = name[0]
    forename = ""
    initials = ""
    initial = ""
    if len(name) > 1:
        forename = name[1]
        initials = forename[0]
        initial = initials
        if len(name) == 3:
            forename = name[1]
            initials = name[1][0] + name[2][0]
            initial = name[1][0]
    # print ('** lastname, forename, initial, initials', lastname, forename, initial, initials)
    return lastname, forename, initial, initials


def searchPapers(author, institute="Fred Hutch"):
    if institute:
        query = "%s*[Affiliation] AND %s[Author]" % (institute, author)
    else:
        query = "%s[Author]" % author
    results = _EntrezSearch(query)
    id_list = results["IdList"]
    papers = _EntrezDetails(id_list)
    if not papers:
        print("no papers found for %s %s" % (author, institute))
        return []
    return papers


def _EntrezSearch(query):
    from Bio import Entrez

    Entrez.email = "your.email@example.com"
    handle = Entrez.esearch(
        db="pubmed", sort="relevance", retmax="2000", retmode="xml", term=query
    )
    results = Entrez.read(handle)
    return results


def _EntrezDetails(id_list):
    from Bio import Entrez

    ids = ",".join(id_list)
    try:
        Entrez.email = "your.email@example.com"
        handle = Entrez.efetch(db="pubmed", retmode="xml", id=ids)
        results = Entrez.read(handle)
        return results
    except:
        return []


def _month2num(month):
    d = {
        "January": "01",
        "February": "02",
        "March": "03",
        "April": "04",
        "May": "05",
        "June": "06",
        "July": "07",
        "August": "08",
        "September": "09",
        "October": "10",
        "November": "11",
        "December": "12",
        "Jan": "01",
        "Feb": "02",
        "Mar": "03",
        "Apr": "04",
        "May": "05",
        "Jun": "06",
        "Jul": "07",
        "Aug": "08",
        "Sep": "09",
        "Oct": "10",
        "Nov": "11",
        "Dec": "12",
    }
    if month in d:
        return d[month]
    else:
        return month




"""
The sci package contains helper and wrapper functions
for scientists  
"""
__version__ = "0.1.3"

#from sci.store import *
#from sci.db import *
#from sci.compute import *
#from sci.json import *
#from sci.new import *
#from sci.old import *
#from sci.tools import *

import sci.compute
import sci.store
import sci.db
import sci.fh
import sci.json
import sci.tools
import sci.old
import sci.new



# ### sci-pkg ###
# json: functions that handle json files
#

import json

def jcompare(jsonfile_or_list,newlist):
    """ eg: addedlist, removedlist = jcompare('/tmp/cache.json',['a', 'b'])
    compares previously saved (json) list with a new list and returns
    a list of newly added items and a list of removed items.
    The old list can be a Python list or a json list stored in the file system
    """
    import os
    oldlist = []
    addedlist, removedlist = newlist, []
    if isinstance(jsonfile_or_list,list):
        oldlist=jsonfile_or_list
    elif os.path.exists(jsonfile_or_list):
        with open(jsonfile_or_list, 'r') as f:
            oldlist=json.load(f)

    addedlist = [item for item in newlist if item not in oldlist]
    removedlist = [item for item in oldlist if item not in newlist]
    return addedlist, removedlist

def jsearch(json,sfld,search,rfld):
    """ return a list of values from a column based on a search """
    lst=[]
    for j in json:
        if j[sfld]==search or search == '*':
            lst.append(j[rfld].strip())
    return lst

def jgetonerow(json,sfld,search):
    """ return a row based  on a search """
    for row in json:
        if row[sfld]==search or search == '*':
            return row

def jsearchone(json,sfld,search,rfld):
    """ return the first search result of a column based search """
    for j in json:
        if j[sfld]==search:
            return j[rfld].strip()

def jget(json,rfld):
    """ return all values in one column """
    lst=[]
    for j in json:
        if j[rfld].strip() != "":
            lst.append(j[rfld].strip())
    return lst



import os, sys, subprocess, atexit
from setuptools import setup
from setuptools.command.install import install

__version__ = "0.1.8"

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Unix Shell",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Systems Administration",
    "Topic :: Utilities"
]
OSPKG_APT=['fping','unixodbc-dev']
OSPKG_YUM=['fping','unixODBC-devel']

class CustomInstall(install):
    def run(self):
        def _post_install():
            pass
            #def find_module_path():
            #    for p in sys.path:
            #        if os.path.isdir(p) and my_name in os.listdir(p):
            #            return os.path.join(p, my_name)
            #install_path = find_module_path()
            #print('    package installed in %s' % install_path)

        # Add your post install code here
        if os.path.exists('/usr/bin/apt-get'):
            inst=['apt-get', 'install', '-y']
            print('    executing %s ...' % " ".join(inst+OSPKG_APT)) 
            subprocess.Popen(inst+OSPKG_APT)
        elif os.path.exists('/usr/bin/yum'):
            inst=['yum', 'install']
            print('    executing %s ...' % " ".join(inst+OSPKG_YUM))
            subprocess.Popen(inst+OSPKG_YUM)

        atexit.register(_post_install)
        install.run(self)

setup(
    name='Omar8few4',
    version=__version__,
    description='a collection of convenience and wrapper functions supporting tasks frequently needed by scientists',
    long_description=open('README.rst', 'r').read(),
    packages=['sci'],
    #scripts=[
    #    'bin/sci',
    #    ],
    author = 'Dirk Petersen',
    author_email = 'dp@nowhere.com',
    url = 'https://github.com/FredHutch/sci-package',
    download_url = 'https://github.com/FredHutch/sci-package/tarball/%s' % __version__,
    keywords = ['hpc', 'scientific computing'], # arbitrary keywords
    classifiers = CLASSIFIERS,
    install_requires=[
        'requests',
        'python-hostlist',
        'psycopg2-binary',
        'PyMySQL',
        'pyodbc',
        'python-swiftclient',
        'python-keystoneclient',
        ],
    cmdclass={'install': CustomInstall},
    #entry_points={
    #    # we use console_scripts here to allow virtualenv to rewrite shebangs
    #    # to point to appropriate python and allow experimental python 2.X
    #    # support.
    #    'console_scripts': [
    #        'slurm-script=slurm_toys.slurm_script:main',
    #    ]
    #}
)
